/**
 * @file FFIntercept.cc
 * Implementation of the FFIntercept skill.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2001/09/29 18:55:20 $
 */

#include "Line.h"
#include "FFIntercept.h"
#include "KickCommand.h"
#include "TurnNeckCommand.h"
#include "KickBallToSpeedVector.h"
#include "TAngle.h"

RS_BEGIN_NAMESPACE

  // #define destDir 0
  //agdir
//  #define BODY_DIST 0.5
//  #define BACK_ANGLE 130
//  #define FRONT_ANGLE 55

// --- CONSTRUCTORS ---

/**
 * Constructor.
 */

FFIntercept::FFIntercept(int maxStp)
  : Skill(),
    maxSteps(maxStp),
    worldView(NULL),
    parameterView(NULL),
    goalie(false)
{
  getView(worldView, WORLD_STATE_VIEW_ID);
  getView(parameterView, PARAMETER_VIEW_ID);
}

// --- DESTRUCTOR ---

FFIntercept::~FFIntercept()
{
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(PARAMETER_VIEW_ID);
}

// --- METHODS ---
void FFIntercept::setMaxSteps(int steps)
{
  maxSteps = steps;
}

void FFIntercept::setGoalie(bool g)
{
  goalie = g;
}

bool FFIntercept::getGoalie()
{
  return goalie;
}

/**
 * Genrates a plan for achieving a pass.
 * Chooses the closest of the passvictims if the argument is
 * supplied, otherwise pass the best player according to 
 * TerritoryControlView
 */

bool FFIntercept::goalFound(const WorldState& ws)
{
  if(!goalie)
  {
    // Goal state for non goalie players.
    TVector delta = 
      ws.getBall().getPosition()
      - ws.getAgent().getPosition();
    if(delta.getLength() < parameterView->SP_kickable_area * 0.8)
    {
      //      cout << "Goal found, balldist: " << delta.getLength() << endl;
      return true;
    }
  }
  else
  {
    // Goal state for goalie player.
    TVector delta = 
      ws.getBall().getPosition()
      - ws.getAgent().getPosition();
    if( delta * ws.getAgent().getDirectionVector() > 0.0
	&& delta.getLength() < parameterView->SP_catchable_area_l )
    {
      return true;
    }
  }

  return false;
}

bool FFIntercept::opTurn(WorldState& ws, const BallState& ball, Step& opStep)
{
  TVector aToB = ball.getPosition()
    - ws.getAgent().getPosition();

  // Try to get in front of the ball.
  //  aToB = aToB + 0.2 * (TVector::TPolarVector(1, ball.getSpeed().getAngle()));

  // Remove the agent's speed.
  int deltaT = ball.getGameTime() - ws.getGameTime();
  double scale = 0.0;

  // FIXME: Table or something.
  for(int i = 0; i < deltaT; i++)
  {
    scale += pow(parameterView->SP_player_decay, i);
  }

  TVector agentSlide = ws.getAgent().getSpeed() * scale;
  aToB -= agentSlide;

  TAngle diff = aToB.getAngle() - ws.getAgent().getDirection();
  bool result = ws.doActualTurn(diff);
  ws.doTurnNeckTowardsBall();
  opStep = ws.getStep();
  ws.step();

//    AgentState tmpAgent = ws.getAgent();
//    for(int i = 0; i < deltaT-1; i++)
//    {
//      double dist = (tmpAgent.getPosition() - ball.getPosition()).getLength();
//      tmpAgent.doDashDistance(dist);
//      tmpAgent.step();
//    }
//    TVector delta = tmpAgent.getPosition() - ball.getPosition();
//    cout << "dlt: " << delta.getLength()<< endl;

  // FIXME: Temporary (or permanent??) hack.
  BallState bl = ws.getBall();
  bl.setC(1.0);
  ws.setBall(bl);

  AgentState agent = ws.getAgent();
  agent.setC(1.0);
  ws.setAgent(agent);

  return result;
}

bool FFIntercept::opDash(WorldState& ws, Step& opStep)
{
  // Look at the next ballposition
  BallState nextBall = ws.getBall();
  nextBall.setAcceleration(TVector(0,0));
  nextBall.step();

  TLine bLine = TLine(ws.getBall().getPosition(),
		      ws.getBall().getPosition() + ws.getBall().getSpeed());

  TLine pLine = TLine(ws.getAgent().getPosition(),
		      ws.getAgent().getPosition() 
		      + ws.getAgent().getDirectionVector());

  double needDist = 0.0;
//    //    if(ws.getAgent().getSpeed().getLength() < 0.05
//    //       || ws.getBall().getSpeed().getLength() < 0.05)
//    //      needDist = (ws.getBall().getPosition() - ws.getAgent().getPosition()).getLength();
//    //    else


  TVector ipos = ws.getBall().getPosition();
  if(ws.getBall().getSpeed().getLength() > 0.1)
    ipos = pLine.getIntersection(bLine);

  needDist = ( ipos -
  	      ws.getAgent().getPosition()).getLength();

  // Goalies wants to intercept the ball at catchable_area_l/2
  if(goalie)
    needDist = needDist - parameterView->SP_catchable_area_l / 2.0;

  bool result = ws.doDashDistance(needDist);

  // Hard code out the back-intercepts.
  if(needDist < -0.4)
    result = false;

  //  bool result = ws.doDashDistance(0.05);
  //  ws.doTurnNeckTo(lookDir);
  ws.doTurnNeckTowardsBall();
  opStep = ws.getStep();
  ws.step();

  // FIXME
  // UGLY!
  BallState bl = ws.getBall();
  bl.setC(1.0);
  ws.setBall(bl);

  AgentState agent = ws.getAgent();
  agent.setC(1.0);
  ws.setAgent(agent);

  return result;
}

Plan FFIntercept::generatePlan()
{
  // Check if the cached version is uptodate.
  if(planTime == worldView->getWorldState().getGameTime())
    return thePlan;

  // Cached plan is invalid, generate a new plan.
  planTime = worldView->getWorldState().getGameTime();

  // Two cases.
  // a) dash-dash-dash
  // b) Turn-to-ball, dash-dash-dash
  WorldState curWs = worldView->getWorldState();
  WorldState nextWs;
  curWs.reduceCR(0.3, 20);

//    if(goalFound(curWs))
//    {
//      cout << "Why intercept when the goal has been achieved?" << endl;
//      return Plan();
//    }

  // Evaluate dash-dash

  Step tmp;

  nextWs = curWs;
  int ddSteps = 10000;
  Step ddStep;
  //    cout << "Distances: ";
  for(int i = 0; i < maxSteps; i++)
  {
    //        TVector delta = 
    //  	nextWs.getBall().getPosition() 
    //  	- nextWs.getAgent().getPosition();
    //        cout << "(" << i << " - " << delta.getLength() << ")" << endl;
    if(goalFound(nextWs))
    {
      //	  cout << "dd-nextWs.gt" << nextWs.getGameTime() << endl;
      ddSteps = i;
      //	  cout << "goalFound - ddSteps:" << ddSteps << endl;
      break;
    }
    if(i == 0)
    {
      // First step is always a dash.
      opDash(nextWs, ddStep);
    }
    else
    {
      // The others are, erm, well.. dashes...
      opDash(nextWs, tmp);
    }
  }
  
  //    cout << endl;

  // Evaluate turn-dash
  int tdSteps = 10000;
  Step tdStep;
  for(int i = 0; i < maxSteps; i++)
  {
    Step curtdStep;
    int curtdSteps = 10000;
    nextWs = curWs;
      
    BallState endBall = curWs.getBall();
    AgentState endAgent = curWs.getAgent();
    int endStep = i*1.2;
    for(int j = 0; j < endStep; j++)
    {
      endBall.setAcceleration(TVector(0,0));
      endBall.step();
      endAgent.doNull();
      endAgent.step();
    }

    TVector delta = endBall.getPosition() - endAgent.getPosition();
    if(delta.getLength() 
       < parameterView->SP_player_speed_max * i)
    {
      for(int j = 0; j < i; j++)
      {
	// Have we found the goal yet?
	if(goalFound(nextWs))
	{
	  curtdSteps = j;
	  //  	      cout << "goalFound - tdSteps:" << curtdSteps << endl;
	  //  	      cout << "td-nextWs.gt" << nextWs.getGameTime()
	  //  		   << " endBall.gt" << endBall.getGameTime() << endl;
//  	  if(j == 0)
//  	    cout << "Strange! Goal found before dashing in intercept" 
//  		 << endl;
	  break;
	}

	if(j == 0)
	{
	  // First cycle is always a turn
	  opTurn(nextWs, endBall, curtdStep);
	}
	else
	{
	  // All others are dashes.
	  opDash(nextWs, tmp);
	}
      }

      // If a goal is found it is the optimal one.
      if(curtdSteps < 1000)
      {
	tdSteps = curtdSteps;
	tdStep = curtdStep;
	break;
      }
    }
  }
  //    cout << "dd: " << ddSteps
  //         << " td: " << tdSteps << endl;

  if(ddSteps > 1000
     && tdSteps > 1000)
  {
    possible = false;
    thePlan = Plan();
    return thePlan;
  }

  possible = true;
  if(tdSteps - ddSteps > -5) // Allow a 2 cycle delay to avoid turns.
  {
//      cout << "I(" << curWs.getGameTime() << "): dash" << endl;
    thePlan = Plan(ddStep);
    return thePlan;
  }

//    cout << "I(" << curWs.getGameTime() << "): turn" << endl;
  thePlan = Plan(tdStep);
  return thePlan;
}

/**
 * Persistent.
 * FALSE
 */
FuzzyBool FFIntercept::persistent()
{
  return FuzzyBool(0.0);
}

/**
 * Returns the applicable value
 */

FuzzyBool FFIntercept::applicable()
{
  // Already at the goal state, why run the skill?
//    if(goalFound(worldView->getWorldState()))
//      return FuzzyBool(0.0);

  if(worldView->getWorldState().getBall().getC() < 0.5)
    return FuzzyBool(0.0);

  generatePlan();

  if(possible)
    return FuzzyBool(1.0);
  return FuzzyBool(0.0);
}

FuzzyBool FFIntercept::succeed()
{
  return FuzzyBool(1.0);
}

FuzzyBool FFIntercept::failed()
{
  return FuzzyBool(0.0);
}

RS_END_NAMESPACE

