/* -*- Mode: C++ -*-
 * $Id: FFScore.cc,v 1.3 2001/09/29 18:55:20 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *0
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: FFScore.cc
 * Unitname: Framework/skill
 * $Revision: 1.3 $
 * Created by:Chris & Govert
 * Last modified by $Author: frehe $ $Date: 2001/09/29 18:55:20 $
 *
 *
 * HISTORY
 */

/**
 * @file FFScore.cc
 * Implementation file for the skill class FFScore.
 *
 * @par Unit:
 *      Framework/skill
 * $Revision: 1.3 $
 * $Author: frehe $
 * $Date: 2001/09/29 18:55:20 $
 */


#include "FFScore.h"
#include "ViewManager.h"
#include "AngleThingy.h"
#include <list>
#include <algorithm>

RS_BEGIN_NAMESPACE

/**
 * This class describes the clearball skill. It's called in an emergency (probably 
 * in front of the own goal) and it results in kicking the ball to an open area on 
 * the field (preferably towards the enemy side of the field)
 *
 * Preconditions
 *  The player needs to be in kicking distance to the ball
 *  It is impossible to give a good pass to a friendly player
 *  It is to dangerous to try to dribble (e.g. if the enemy players get the ball they will score)
 * Postconditions
 *  The ball is kicked to an empty area on the field if an appropriate area exists.
 *  if no good area exists within the view of the player he tries to kick it in the direction 
 *  of the enemy goal
 *  the ball is NEVER kicked in the direction of the own side of the field (because this could 
 *  cause a more dangerous situation)
 *  The ball is only kicked outside of the field if there is no better solution
 */
FFScore::FFScore()
  : Skill(),
    possible(false),
    calculatedTime(-1),
    thePlan(),
    worldView(NULL)
{
  getView(worldView, WORLD_STATE_VIEW_ID);
  getView(parameterView, PARAMETER_VIEW_ID);
} 

FFScore::~FFScore()
{
  viewManager->releaseView(WORLD_STATE_VIEW_ID);
  viewManager->releaseView(PARAMETER_VIEW_ID);
}

//------------------------------------------------------------

Plan FFScore::generatePlan()
{
  if(calculatedTime == worldView->getWorldState().getGameTime())
  {
    return thePlan;
  }
  calculatedTime = worldView->getWorldState().getGameTime();
  possible = false;

  // Construct an anglethingy.
  AngleThingy angles;

  // Initialize anglethingy.
  // Assume max kick.
  const AgentState& agent
    = worldView->getWorldState().getAgent();
  BallState ball
    = worldView->getWorldState().getBall();

  double kpr = agent.getKickPowerRate(ball);
  double v0 = kpr * 100.0;
//    angles.setInitialSpeed(v0);
  angles.setInitialSpeed(v0);

  // Find the initial span, the span between the two goalposts in this case.
  TVector leftLimit = 
    TVector(parameterView->SP_pitch_length/2,
	    -0.95*parameterView->SP_goal_width/2) - ball.getPosition();
  TVector rightLimit = 
    TVector(parameterView->SP_pitch_length/2,
	    0.95*parameterView->SP_goal_width/2) - ball.getPosition();

  AngleThingy::Span initSpan;
  initSpan.start = leftLimit.getAngle();
  initSpan.end = rightLimit.getAngle();
  angles.addSpan(initSpan);

  // Time for heavy black magic.
  // Split the angles
  float bestDir = 0.0;
  float bestSpan = -1.0;
  TLine goalLine(TVector(parameterView->SP_pitch_length/2,
			 -parameterView->SP_goal_width/2),
		 TVector(parameterView->SP_pitch_length/2,
			 parameterView->SP_goal_width/2));

  const int maxSteps = 40;
  int time;
  for(time = 0; time < maxSteps && angles.getSpans().size() > 0; time++)
  {
    angles.step();
    const std::list<AngleThingy::Span>& spans = angles.getSpans();
    std::list<AngleThingy::Span>::const_iterator itSpans;

//      if(spans.size() == 0)
//        break;

    for(itSpans = spans.begin(); itSpans != spans.end(); itSpans++)
    {
      if(time != 0 && (itSpans->end - itSpans->start) > bestSpan )
      {
  	const BallState& ball = angles.getBall();
	float ballRadius = angles.getBallRadius();
	TLine leftLine(ball.getPosition(),
		       ball.getPosition()
		       +TVector::TPolarVector(ballRadius, itSpans->start));
	TLine rightLine(ball.getPosition(),
			ball.getPosition()
			+TVector::TPolarVector(ballRadius, itSpans->end));
	double s1, t1, s2, t2;
	TVector tmp = goalLine.getIntersection(leftLine, t1, s1);
	tmp = goalLine.getIntersection(rightLine, t2, s2);
	if(s1 < 1.0 && s2 < 1.0)
	{
	  bestSpan = itSpans->end - itSpans->start;
	  bestDir = (itSpans->end + itSpans->start) / 2.0;
	  possible = true;
	}
      }
    }
  }

//    if(possible)
//    {
//      cout << "Score possible, dir: " << bestDir << endl;
//    }
//    else
//    {
//      cout << "Score impossible" << endl;
//    }

//    TAngle relDir = bestDir 
//      - worldView->getWorldState().getAgent().getDirection();
//    TVector kickVector = TVector::TPolarVector(100.0, relDir);
//    WorldState ws = worldView->getWorldState();
//    ws.doKick(kickVector);

  TVector kickVector = TVector::TPolarVector(3.0, bestDir);
  WorldState ws = worldView->getWorldState();
  ws.doKickToSpeed(kickVector);
  thePlan= Plan(ws.getStep());
  return thePlan;
}

FuzzyBool FFScore::persistent()
{
  return FuzzyBool(0.0);
}


FuzzyBool FFScore::applicable()
{
  const WorldState& ws = worldView->getWorldState();

  if(!ws.getAgent().isBallKickable(ws.getBall()))
    return FuzzyBool(0.0);

  if((ws.getBall().getPosition() 
      - TVector(52.5, 0)).getLength() < 40)
  {
    generatePlan();
    if(possible)
    {
      return FuzzyBool(1.0);
    }
  }
  return FuzzyBool(0.0);
}



FuzzyBool FFScore::succeed()
{
  return FuzzyBool(1.0);
}

FuzzyBool FFScore::failed()
{
  return FuzzyBool(0.0);
}


RS_END_NAMESPACE
