/* -*- Mode: C++ -*-
 * $Id: TeamDecision.cc,v 1.4 2001/09/26 13:02:43 frehe Exp $
 *
 *
 * COPYRIGHT INFORMATION
 *
 * This file is part of RoboSoc created by Fredrik Heintz <frehe@ida.liu.se>
 * Copyright (C) 1999, 2000 Fredrik Heintz, Linkping University, Sweden
 *
 * You are allowed to modify and use this code as long as you retain this
 * notice. If you make any changes or have any comments I would appreciate
 * it if you send me a message. For more information, please see
 * http://www.ida.liu.se/~frehe/RoboCup/RoboSoc/
 *
 *
 * IDENTIFICATION
 *
 * Filename: TeamDecision.cc
 * Unitname: Team
 * $Revision: 1.4 $
 * Created by: Fredrik Heintz 2000-09-16
 * Last modified by $Author: frehe $ $Date: 2001/09/26 13:02:43 $
 *
 *
 * HISTORY
 *
 * $Log: TeamDecision.cc,v $
 * Revision 1.4  2001/09/26 13:02:43  frehe
 * Changed the order of include files to make circular updates less likely.
 *
 * Revision 1.3  2001/09/01 23:54:07  frehe
 * CP_version -> SP_version
 *
 * Revision 1.2  2001/09/01 14:34:06  frehe
 * Updated the template team for the new version of RoboSoc.
 *
 * Revision 1.1  2000/09/16 18:37:03  frehe
 * Add a team that can be used as the basis for new stand alone teams.
 *
 */

/**
 * @file TeamDecision.cc
 * Implementation of the class TeamDecision.
 *
 * @par Unit:
 *      Team
 * $Revision: 1.4 $
 * $Author: frehe $
 * $Date: 2001/09/26 13:02:43 $
 */


#include "TeamDecision.h"

#include <unistd.h>
#include <errno.h>

#include "ViewManager.h"
#include "ParameterView.h"
#include "GameView.h"
#include "CommandView.h"
#include "MarkerView.h"
#include "AgentView.h"
#include "BallView.h"
#include "PlayerView.h"
#include "view_constants.h"
#include "ActuatorInterface.h"
#include "InitCommand.h"
#include "MoveCommand.h"
#include "DashCommand.h"
#include "KickCommand.h"
#include "TurnCommand.h"
#include "TurnNeckCommand.h"
#include "SayCommand.h"
#include "FindBall.h"
#include "GotoBallAvoidingObstacles.h"
#include "MoveToPositionWithBall.h"
#include "Score.h"
#include "rs_assert.h"
#include "rs_log.h"


RS_BEGIN_NAMESPACE


TeamDecision::TeamDecision(ActuatorInterface* const actuator_interface)
  : Decision(actuator_interface),
    viewManager(ViewManager::instance()),
    parameterView(NULL),
    gameView(NULL),
    commandView(NULL),
    markerView(NULL),
    ballView(NULL),
    agentView(NULL),
    playerView(NULL),
    serverAlive(true)
#ifdef RS_USE_DISPLAY
  , theDisplay()
#endif
#ifdef RS_USE_COACH_DEBUGGING
  , coachView(NULL)
#endif
#ifdef RS_USE_OWN_LOGFILE
  , logFile("Team.log")
#endif
{
  getView(parameterView, PARAMETER_VIEW_ID);
  getView(gameView, GAME_VIEW_ID);
  getView(commandView, COMMAND_VIEW_ID);
  getView(markerView, MARKER_VIEW_ID, parameterView);
  getView(ballView, BALL_VIEW_ID);
  getView(agentView, AGENT_VIEW_ID, parameterView);
  getView(playerView, PLAYER_VIEW_ID);
  
#ifdef RS_USE_COACH_DEBUGGING
  getView(coachView, COACH_VIEW_ID);
#endif
}


TeamDecision::~TeamDecision()
{
  viewManager->releaseView(PARAMETER_VIEW_ID);
  viewManager->releaseView(GAME_VIEW_ID);
  viewManager->releaseView(COMMAND_VIEW_ID);
  viewManager->releaseView(MARKER_VIEW_ID);
  viewManager->releaseView(BALL_VIEW_ID);
  viewManager->releaseView(AGENT_VIEW_ID);
  viewManager->releaseView(PLAYER_VIEW_ID);
#ifdef RS_USE_COACH_DEBUGGING
  viewManager->releaseView(COACH_VIEW_ID);
#endif
}


ServerCommand* TeamDecision::initAgent()
{
  RS_REQUIRE( parameterView != NULL );
  InitCommand* init_cmd = new InitCommand(parameterView->CP_team_name,
					  parameterView->SP_version,
					  parameterView->CP_goalie);
  return init_cmd;
}


bool TeamDecision::start()
{
  RS_REQUIRE( initialized == true );
  try {
    while ( serverAlive )
      try {
	if ( pause() == -1 && errno != EINTR )
	  RS_ERROR("Something went wrong in pause!");
      } catch (RSException &excpt) {
	std::cerr << "Caught a RoboSoc exception: " << excpt.What() << std::endl;
	return 1;
      } catch (std::exception &excpt) {
	std::cerr << "Caught a standard exception: " << excpt.what() << std::endl;
	return 1;
      } catch (...) {
	std::cerr << "Caught an unknown exception: " << std::endl;
	return 1;
      }
    return true;
  } catch (RSException &excpt) {
    std::cerr << "Caught a RoboSoc exception: " << excpt.What() << std::endl;
    return 1;
  } catch (std::exception &excpt) {
    std::cerr << "Caught a standard exception: " << excpt.what() << std::endl;
    return 1;
  } catch (...) {
    std::cerr << "Caught an unknown exception: " << std::endl;
    return 1;
  }
}


bool TeamDecision::decide()
{
  static FindBall findBall;
  static GotoBallAvoidingObstacles gotoBallAvoidingObstacles;
  static MoveToPositionWithBall moveToPositionWithBall(Point(55,0));
  static Score score;

  Plan plan;

  if ( score.applicable() ) {
    plan = score.generatePlan();
  }
  else if ( moveToPositionWithBall.applicable() ) {
    plan = moveToPositionWithBall.generatePlan();
  }
  else if ( gotoBallAvoidingObstacles.applicable() ) {
    plan = gotoBallAvoidingObstacles.generatePlan();
  }
  else {
    plan = findBall.generatePlan();
  }

  actuatorInterface->addToPlan(plan.getNextStep());

  
  return true;
}


void TeamDecision::onActuatorSensorData()
{
#ifdef RS_USE_DISPLAY
  theDisplay.setLastSkill(commandView->getLastServerString());
#endif
#ifdef RS_USE_OWN_LOGFILE
  logFile << gameView->getCurrentTime() << ": executed action "
	  << commandView->getLastServerString()
	  << std::endl;
  if ( commandView->getActualTurn(NOW) > AngleDeg(0) )
    logFile << gameView->getCurrentTime() << ": actual turn = "
	    << commandView->getActualTurn(NOW)
	    << std::endl;  
#endif
}


void TeamDecision::onInit()
{
  MoveCommand* move_cmd = new MoveCommand(Point(-10, -10));
  actuatorInterface->addToPlan(move_cmd);
}


void TeamDecision::onVisualSensorData()
{
  //RS_LOG( LA_DECISION, gameView->getCurrentTime() << ": **saw**" );
}


void TeamDecision::onAfterSensorData()
{
}


void TeamDecision::onNewCycle(const bool forced)
{
  if ( forced ) {
    RS_LOG( LA_DECISION,
	    gameView->getCurrentTime() << ": **Forced NewCycle**" );
  }
  else {
    RS_LOG( LA_DECISION, gameView->getCurrentTime() << ": **NewCycle**" );
  }

#ifdef RS_USE_DISPLAY
  theDisplay.draw();
#endif

}


bool TeamDecision::onEstimatedNewCycle(UInt16 estimated_time, UInt16 ticks)
{
  // We assume that if the sense_body is more than four ticks late
  // then it's lost.
  RS_LOG( LA_DECISION, "onEstimatedNewCycle(" << estimated_time
	  << ", " << ticks << ")" );

  // If more than nine sense_body's are missed in a row then assume that
  // the server is dead.
  if ( estimated_time > 9 )
    serverAlive = false;

  if ( ticks > 4 ) {
    // TODO: change this into sending a control command instead!
    //sensorInterface->ForceNewCycle();
    return true;
  } else
    return false;
}


void TeamDecision::onCommandWarning()
{
  decide();
}


RS_END_NAMESPACE
